package mdd.client;


import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.geometry.Bounds;
import javafx.geometry.Point2D;
import javafx.scene.Group;
import javafx.scene.Node;
import javafx.scene.control.ComboBox;
import javafx.scene.control.Label;
import javafx.scene.control.TextField;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.GridPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.scene.shape.Rectangle;
import javafx.scene.text.Font;
import javafx.scene.text.Text;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;

import javax.swing.*;
import java.util.Vector;


public class Module extends Group implements IGUIEventClient, ChangeListener<Bounds> {
    private boolean _dragged = false;
    private Point2D _pos = new Point2D(0.0f,0.0f);
    private Rectangle titlebar;
    private Rectangle background;
    private Text caption;
    protected GUIEventHandler _eventHandler;

    public String Name = "TEST";
    public long Appendix = 0;
    public String ID = "TEST-1";
    public Vector<String> Prefix;

    protected String Type = "module";
    protected String Key = "";

    public VBox vlayout = new VBox();
    GridPane configs = new GridPane();
    public BorderPane hlayout = new BorderPane();
    public VBox inputs = new VBox();
    public VBox outputs = new VBox();
    public boolean isSelected = false;

    public Module(JSONObject json){
        Name = json.get("name").toString();
        Appendix = (long)json.get("id");
        ID = Name +Appendix;
        Type = json.get("type").toString();
        Key = json.get("key").toString();
        Prefix = new Vector<>();
        JSONArray jarr = (JSONArray)json.get("prefix");
        for (Object o : jarr) {
            Prefix.add(o.toString());
        }

        if (json.containsKey("inputs")){
            inputs.getChildren().clear();
            for (Object o : (JSONArray)json.get("inputs")){
                addInput(new Input(this,(JSONObject)o));
            }
        }

        if (json.containsKey("outputs")) {
            outputs.getChildren().clear();
            for (Object o : (JSONArray) json.get("outputs")) {
                addOutput(new Output(this, (JSONObject) o));
            }
        }

        if (json.containsKey("configure")) {
            configs.getChildren().clear();
            configs.setHgap(20);
            configs.setVgap(20);
            int counter = 0;
            for (Object o : (JSONArray) json.get("configure")) {
                JSONObject jconfig = (JSONObject) o;
                HBox hbox = new HBox();
                configs.add(new Label((String) jconfig.get("name")),0,counter);
                if (jconfig.containsKey("options")){
                    ObservableList<String> options = FXCollections.observableArrayList();
                    System.out.println(jconfig.toString());
                    for (Object opt : (JSONArray) jconfig.get("options")){
                        String item = opt.toString();
                        options.add(opt.toString());

                    }
                    ComboBox cbox = new ComboBox(options);
                    cbox.getSelectionModel().select(jconfig.get("value").toString());
                    configs.add(cbox,1,counter);
                }
                else{
                    configs.add(new TextField(jconfig.get("value").toString()),1,counter);
                }
                configs.getChildren().add(hbox);
                ++counter;
            }
        }
        if (json.containsKey("GUI")) {
            JSONObject jgui = (JSONObject) json.get("GUI");
            setTranslateX((double)jgui.get("posX"));
            setTranslateY((double)jgui.get("posY"));
        }

        Color color = Color.DARKCYAN;

        setFocusTraversable(true);
        _eventHandler = GUIEventHandler.getEventHandler();
        _eventHandler.addEventListener(this);

        inputs.setSpacing(10);

        outputs.setSpacing(10);

        hlayout.setLeft(inputs);
        hlayout.setRight(outputs);
        //hlayout.getChildren().addAll(inputs, outputs);
        //hlayout.setSpacing(200);
        inputs.setTranslateX(-13);
        outputs.setTranslateX(13);
        vlayout.getChildren().addAll(configs, hlayout);
        vlayout.setTranslateX(5);
        vlayout.setTranslateY(30);

        // System.out.println("[Module]: pressed: "+ hbox.getPrefWidth());


        //titlebar = new Rectangle(0.0f, 0.0f, 150-20, 20.0f);
        titlebar = new Rectangle(0.0f, 0.0f, 1.0, 20f);
        titlebar.setArcWidth(10.0f);
        titlebar.setArcHeight(10.0f);
        titlebar.setFill(color);
        //ackground = new Rectangle(0.0f, 0.0f, 150-20, 150 + 40);
        background = new Rectangle(0.0f, 0.0f, 1.0, 1.0);
        background.setArcWidth(10.0f);
        background.setArcHeight(10.0f);
        background.setFill(Color.GRAY);
        caption = new Text(Name);
        caption.setFont(new Font("Arial", 15));
        caption.getStyleClass().add("text");
        caption.setX(10);
        caption.setY(15);


        getChildren().addAll( background, titlebar, caption, vlayout);



        addEventFilter(MouseEvent.MOUSE_PRESSED, e -> {
            if(e.getButton() == MouseButton.PRIMARY) {
                isSelected = true;
                System.out.println("[Module]: select: " + ID);
                _eventHandler.publishEvent("MOUSE_PRESSED", getJSONIdentifier());
                _pos = new Point2D(e.getSceneX(), e.getSceneY());
            }
            if(e.getButton() == MouseButton.SECONDARY) {

            }
            //e.consume();
        });

        addEventHandler(MouseEvent.MOUSE_DRAGGED, e -> {
            // System.out.println("[Module]: dragg event detected!");
            if(e.getButton() == MouseButton.PRIMARY) {
                if (isSelected){
                    setTranslateX(getTranslateX()+(e.getSceneX()-_pos.getX()));
                    setTranslateY(getTranslateY()+(e.getSceneY()-_pos.getY()));
                    _pos = new Point2D( e.getSceneX(),  e.getSceneY());
                }
            }
            if(e.getButton() == MouseButton.SECONDARY) {

            }
        });

        addEventHandler(MouseEvent.MOUSE_RELEASED, e -> {
            // System.out.println("[Module]: dragg event detected!");
            if(e.getButton() == MouseButton.PRIMARY) {
                if (isSelected){
                    JSONObject jgui_args = new JSONObject();
                    jgui_args.put("posX", getTranslateX());
                    jgui_args.put("posY", getTranslateY());
                    JSONObject jgui = new JSONObject();
                    jgui.put("GUI", jgui_args);
                    JSONObject jmsg = new JSONObject();
                    jmsg.put("operation", "change");
                    JSONObject jargs = new JSONObject();
                    jargs.put("subject",  getJSONIdentifier());
                    jargs.put("object",   jgui);
                    jmsg.put("args", jargs);
                    _eventHandler.publishEvent("try", jmsg);
                }
            }
            if(e.getButton() == MouseButton.SECONDARY) {

            }
        });

        vlayout.layoutBoundsProperty().addListener(this);
    }

    public void addInput(Input in){
        inputs.getChildren().add(in);
    }

    public void addOutput(Output out){
        outputs.getChildren().add(out);
    }

    public Input getInput(String id){
        for (Node node : inputs.getChildren()){
            if(node instanceof Input){
                Input in = (Input) node;
                if (in.ID.equals(id)){
                    return in;
                }
            }

        }
        System.out.println("[Module]: getInput: WRONG ID!");
        return null;
    }

    public Output getOutput(String id){
        for (Node node : outputs.getChildren()){
            if(node instanceof Output){
                Output out = (Output) node;
                if (out.ID.equals(id)){
                    return out;
                }
            }

        }
        System.out.println("[Module]: getOutput: WRONG ID!");
        return null;
    }

    public JSONObject getJSONIdentifier(){
        JSONObject json = new JSONObject();
        json.put("prefix", Prefix);
        json.put("name", Name);
        json.put("id", Appendix);
        json.put("type", Type);
        json.put("key", Key);
        return json;
    }

    public Vector<String> getAsPrefix(){
        Vector<String> ret = (Vector<String>)Prefix.clone();
        ret.add(ID);
        return ret;
    }

    @Override
    public void processGUIEvent(String event, JSONObject args) {
        if (event == "MOUSE_PRESSED"){
            if (!args.equals(getJSONIdentifier())){
                isSelected = false;
                return;
            }
            return;
        }
        if (event.equals("change")){
            if (!args.containsKey("args")){
                return;
            }
            JSONObject jargs = (JSONObject) args.get("args");
            if (!jargs.containsKey("subject")){
                return;
            }
            JSONObject jsub = (JSONObject) jargs.get("subject");
            if (!jsub.equals(getJSONIdentifier())){
                return;
            }
            if (!args.containsKey("operation")){
                return;
            }
            JSONObject jobj = (JSONObject) jargs.get("object");
            switch((String) args.get("operation")){
                case "change":
                    if (jobj.containsKey("GUI")){
                        JSONObject jgui = (JSONObject) jobj.get("GUI");
                        setTranslateX((double)jgui.get("posX"));
                        setTranslateY((double)jgui.get("posY"));
                    }
            }
            return;
        }
    }

    @Override
    public void close() throws Exception {
        _eventHandler.removeEventListener(this);
    }

    @Override
    public void changed(ObservableValue<? extends Bounds> observable, Bounds oldValue, Bounds newValue) {
        titlebar.setWidth(newValue.getWidth()+10);
        titlebar.setHeight(20.0f);
        background.setWidth(newValue.getWidth()+10);
        background.setHeight(newValue.getHeight()+vlayout.getTranslateY()+10);
        Bounds hlayout_bounds = hlayout.getLayoutBounds();
        //double val = newValue.getWidth()-hlayout.getWidth()-hlayout.getSpacing();
        //System.out.println("[Module]: test" + val+ " " + hlayout.getWidth());
    }
}
